package org.vishia.byteData;

import java.io.UnsupportedEncodingException;

/**This class is a base class to control the access to binary data.
 * The binary data may typically used or produced from a part of software written in C or C++.
 * There the binary data are struct-constructs.
 * <br>
 * It is able to support several kinds of struct constructs:<ul>
 * <li>Simple <code>struct</code> are adequate mapped with a derivated class of this class, 
 *     using the protected commonly access methods like {@link _getInt(int, int)} with predefined indexes 
 *     in special methods like getValueXyz().</li>
 * <li>Complex <code>struct</code> with nested <code>struct</code> inside are mapped 
 *     with one derivated class per <code>struct</code>, define one reference per nested struct 
 *     and overwriting the method {@link assignDataToFixChilds()}</li>
 * <li>Base <code>struct</code> inside a <code>struct</code> (inherition in C) can be mapped with 
 *     extra derivated classes for the base struct and usind the
 *     {@link assignCasted_i(ByteDataAccess, int)}-method.</li>
 * <li>A stacked construct of data with several struct may be mapped using the {@addChild(ByteDataAccess)}-method.
 *     Here a parent should be defined, and the stacked structs are children of this parent. The stacked structs
 *     need not be from the same type.</li>
 * <li>stacked constructs with parent are nestable, it is construable as a tree of stacked structs. 
 *     The parent of the stack is the tree node. It is likewise a XML tree. 
 *     The data may be also transformed to a XML data representation
 *     or the data structure may be explained with a XML tree, but they are not
 *     XML data straight.</li>
 * </ul>
 * This application possibilities show a capable of development system to access binary data. 
 * The other, hand made way was calculate with indices of the byte[idx] specially user programmed. 
 * This class helps to make such complex index calculation needlessly. 
 * One struct at C level corresponds with one derivated class of ByteDataAccess. 
 * But last also a generation of the java code from C/C++-header files containing the structs is able to.
 * So the definition of byte positions are made only on one source. The C/C++ is primary thereby. 
 *      
 * <br>
 * The class pointers a part of data, stored in a byte[]. Some methods are given
 * to access to data, see {@link getInt8(int)} and adequate, or to step to
 * subordinated data of this part of data, named "children" likewise in XML.
 * <br>
 * The base idea is, the controled part of data consists of a head with a fix
 * size and structure, and some children as subordinated data of this part,
 * maybe controled by an other instance of this class or otherwise else.
 * The part of a child need not have a fix size and may have a different structure
 * to its sibling. See the protected method {@link specifyLengthCurrentChildElement()} or
 * public method {@link assignFromParentsChild(XmlBinCodeElement parent)}.
 * <br>
 * In the derivated class the private positions of the elements in the "head"
 * and some public access methods should be defined to simple access to the data.
 * <br>
 * The UML structure of such an class in a environment may be shown with the
 * followed object model diagram, <br>
 * <code> <+>---> </code>is a composition,
 * <code> <>---> </code>is a aggregation, <code> <|---- </code>is a inherition.
 *  <pre>
 *                      +-------------------------------+                 +----------+
 *                      | ByteDataAccess                |----data-------->| byte[]   |
 *                      |-------------------------------|                 +----------+
 *                      |idxBegin:int                   |
 *                      |idxChild:int                   |<---------------+ a known parent
 *  +-------------+     |idxEnd:int                     |---parent-------+ setted in addChild()
 *  | derivated   |     |-------------------------------|
 *  | user        |---|>|specifyLengthElement()         |
 *  | classes     |     |specifyLengthElementHead()     |--currentChild--+ the actual child element
 *  +-------------+     |specifyLengthCurrentChild()    |<---------------+
 *                      +-------------------------------+
 * </pre>
 *
 */

public abstract class ByteDataAccess
{
  /** The array containing the binary data.*/
  protected byte[] data;

  /** Index of the beginning of the actual element in data*/
  protected int idxBegin;

  /** Index of the end of the actual element in data*/
  protected int idxEnd;

  /** Index within the at position of the current child element*/
  protected int idxCurrentChild;

  /** Index used by getText.
   * The idxChildEnd ist incremented inside getText and pointered
   * to the next element. After calling next() this value is load with -1,
   * indicates that no calling of getText() is done. See nextIdx().
   */
  private int idxCurrentChildEnd;

  /** Flag is set or get data in big endian or little endian (if false)*/
  boolean bBigEndian;

  /** The parent XmlBinCodeElement, necessary only for add() and expand().
   *
   */
  private ByteDataAccess parent;


  /** The child on end to add() something
   *
   */
  private ByteDataAccess currentChild;

  /** Definition of the code of end of information, return from next()*/
  public static final byte kEndOfElements = 0;

  /** Definition of the code of no information, return from next()*/
  public static final byte kNothing = (byte)(0xff);

  /** Aussage: es ist ein String (XML: text()), kein Tag im String*/
  public static final byte kText    = 1;

  /** coding: the value is undefined*/
  public static final byte kUndefined  = -0x3f;

  /** Index in the data, position of element code*/
  protected static final int kIdxElementCode = 0;

  /** Conversion of characters in range 0x80..0xbf */
  protected static final String sConvertHiChars = "?^~{}[|]\\.............................................";


  /** Specifies the length of the head data. This is the index of the first child relative to the start position of the element.
   * This method is called inside this base class itself inside some methods. Outside the method may be called
   * at example to calculate the size of data.</br>
   * Only the derivated class knows the position of the first child element.
   * It must specify that like the followed sample:
   * <pre>
   *protected void specifyLengthElementHead()
   *{ return kIdxFirstChild;  //kIdxFistChild should be defined as static final int.
   *}
   *</pre>
   *  @return Position of the first child element relative to the start position or -1.
   */
  public abstract int specifyLengthElementHead();



  /** Sets the elements data to the default empty data.
   * This method should not called outside of this base class, likewise not in the derivated class itself.
   * This method is only called inside this base class itself inside the methods addChild() and assignEmpty().
   * But only the derivated class knows how to set empty data.
   * It must specify that like the followed sample:
   * <pre>
   *protected void specifyEmptyDefaultData()
   *{ data[idxBegin + 0] = 0;
   *  data[idxBegin + kIdxMid] = 0;
   *  data[idxBegin + kIdxMin] = kNotAValue;
   *  data[idxBegin + kIdxMax] = kNotAValue;
   *}
   *</pre>
   */
  abstract protected void specifyEmptyDefaultData();



  /** Returns the actual length of the whole element presenting with this class.
   * This method should be specified in the derivated class.
   * The method is called inside this base class in the methods assignAsChild()
   * to calculate the {@link idxEnd}.
   * <br>
   * The calculation of the whole length of the element should be considered
   * all children inside this element. The structure is user-specific. Only some examples
   * may be given here:
   * <ul><li>If all children have a constant length and the number of children is known,
   *   the calculation of the length is simple: lengthOfHead + nrofChilds * lengthOfChilds.</li>
   * <li>If the children have a variably length, all children should be checked
   *   about its length.<li>
   * </ul>
   * If the actual element are TODO better explain when and why.
   *
   * @return The length of this existing element or -1, if the length is not fix yet.
   * @throws XmlBinCodeElementException If the data inside are corrupted, the user can throw this exception.
   */
  protected abstract int specifyLengthElement()
  throws AccessException;



  /** Returns the length of a child element at current position specified in  the derivated class.
   * The derivated class must known its possible child elements and must get there length
   * with enquiry of specifyLengthElement() of the current child types.
   * This method is called inside the getLengthCurrentChildElement()-method
   * esspecially called inside the next()-method to increment the index idxChild.
   * This method is not called if setLengthCurrentChildElement() was called after calling next().
   * The user may precluding the call of this method if he calls setLengthCurrentChildElement in sequence
   * of the followed sample:
   * <pre>
   * int eChildCode = dataCode.next();
   * switch(eChildCode)
   * { case ...:
   *   { child.assignAsChild(dataCode);             //the child is known yet!
   *     dataCode.setLengthCurrentChildElement(child.getLength());
   *     ....
   *   }
   * <pre>
   * If the user always calls setLengthCurrentChildElement in this manner,
   * he don't need to overwrite specifyLengthCurrentChildElement.
   * @return The length of the existing child element at the current position idxChild.
   */
  protected int specifyLengthCurrentChildElement()
  throws AccessException
  { throw new AccessException("The length of the child is undefined, no user specification is known", this);
  }


  /** Constructs a new emtpy instance. Use assign() to work with it. */
  public ByteDataAccess()
  { this.data = null;
    this.bBigEndian = false;
    idxBegin = 0;
    idxEnd = 0;
    idxCurrentChild = 0;  //to mark start.
    idxCurrentChildEnd = -1;
    parent = null;
    currentChild = null;
  }

  
  
  /**Assigns new data to this element. <br>

   * The user may overwrite this method and call super.assignData(data, length) inside
   * if some additional actions should be done.<br/>
   * This method is also usefull to deassign a current data buffer, call <code>assign(null, 0);</code>.
   * <br>
   * If the element are using before, its connection to an other parent is dissolved.
   * <br>
   * @param data The data. The length of data may be greater as
   *             the number of the significant bytes.
   * @param length Number of significant bytes in data.
   *               If length is > data.length, an exception may be thrown
   *               in any state of the evaluation.
   */
  public void assignData(byte[] data, int length)
  { assignData(data, length, 0);  
  }
  
  
  
  /**Assigns new data to this element at given index in data. <br>
   * The user may overwrite this method and call super.assignData(data, length) inside
   * if some additional actions should be done.<br/>
   * This method is also usefull to deassign a current data buffer, call <code>assign(null, 0);</code>.
   * <br>
   * If the element are using before, its connection to an other parent is dissolved.
   * <br>
   * @param data The data. The length of data may be greater as
   *             the number of the significant bytes.
   * @param length Number of significant bytes in data.
   *               If length is > data.length, an exception may be thrown
   *               in any state of the evaluation.
   * @param index Start position in data 
   */

  public void assignData(byte[] data, int length, int index)
  { this.data = data;
    idxBegin = index; 
    idxEnd = length;
    idxCurrentChild = -1; //no current child is added yet. //specifyLengthElementHead();
    //idxChild = -1;          //next() provides the first element.
    idxCurrentChildEnd = specifyLengthElementHead(); //-1;         //no length of element is known, it means, no child is appended yet.
    if(parent!= null && parent.currentChild == this)
    { parent.currentChild = null;  //the child is invalid because it is new assigned.
    }
    parent = null;
    currentChild = null;
    assignDataToFixChilds();
  }
  
  
  

  

  /**Initializes a top level, the data are considered as non initalized.
   * The length of the head should be a constant value, given from
   * method call {@link specifyLengthElementHead()}. The child Positions
   * are set to the end of head, no childs are presumed.
   * The head should be filled with data after that calling some methods like
   * {@link setInt32(int, int)}.<br>
   * The children should be added by calling {@link addChild(ByteDataAccess)}
   * and filled with data after that.
   *
   * <br>
   * If the element are using before, its connection to an other parent is dissolved.
   * <br>
   * Example to shown the principle:<pre>
   *   ...........the data undefined with defined length.........
   *   +++++                   Head, the length should be known.
   *        ####****#####****  Space for children,
   * </pre>
   * @param data The data. The reference should be initialized, it means
   *        the data have a defined maximum of length. But it is not tested here.
   */
  final public void assignEmpty(byte[] data)
  { this.data = data;
    idxBegin = 0;
    int idxFirstChildElement = specifyLengthElementHead();
    if(idxFirstChildElement <0) { idxCurrentChild = -1; }//there are no childs!
    else                        { idxCurrentChild = idxBegin + idxFirstChildElement; }
    idxEnd = idxCurrentChild;
    idxCurrentChildEnd = -1;
    if(parent!= null && parent.currentChild == this) parent.currentChild = null;
    parent = null;
    currentChild = null;
    specifyEmptyDefaultData();
    assignDataToFixChilds();
  }

  
  
  
  /**Assigns this element to the same position in data, but it is another view. 
   * This method should be called inside a assignCasted() method if a inner head structure is known
   * and the conclusion of this structure is possible. At result, both ByteDataAccess instances reference the same data,
   * in different views.
   * @param dst The known data access
   * @param offsetCastToInput typical 0 if single inherition is used.
   * @throws AccessException if a length of the new type is specified but the byte[]-data are shorter. 
   *                         The length of byte[] is tested. 
   */
  final protected void assignCasted_i(ByteDataAccess dst, int offsetCastToInput, int lengthDst)
  throws AccessException
  { this.data = dst.data;
    this.bBigEndian = dst.bBigEndian;
    if(this.parent!= null && this.parent.currentChild == this) this.parent.currentChild = null;
    this.parent = dst.parent;
    currentChild = null;
    //
    idxBegin = dst.idxBegin + offsetCastToInput;  //to mark start.
    initIdxChildEndFromSpecify(lengthDst);
    //idxEnd = xxx
    assignDataToFixChilds();
  }


  
  
  
  
  /**initializes the idxChild, idxEnd, idxChildEnd from the {@link specifyLengthElementHead()}
   * and {@link specifyLengthElement()} -methods (users derivated).
   * The following elements were initialized:
   * <ul><li>idxChild: from {@link specifyLengthElementHead()}</li>
   * <li>idxChildEnd: set to -1. It means, addChild or next places a new child after head.</li>
   * <li>idxEnd: if {@link specifyLengthElement()} returns a positive value, from it, otherwise set from idxEndP. </li>
   * </ul>
   * The parent should be set before, parent.idxChildEnd is affected if it is the current child.
   * If the length of the element is not specified, the length should be the maximum length of parent.
   * @param idxEndP If the length is given outside and {@link specifyLengthElement()} returns -1 or nothing, this length is used. 
   * @throws AccessException if a length is specified but the byte[]-data are shorter. 
   *                         The length of byte[] is tested. 
   */
  private void initIdxChildEndFromSpecify(int idxEndP) 
  throws AccessException
  { idxCurrentChildEnd = idxBegin + specifyLengthElementHead();
    idxCurrentChild = -1;
    int lengthElement;
    try {lengthElement = specifyLengthElement(); }
    catch(AccessException exc){ lengthElement = -1; }
    if(lengthElement > 0)
    { //the length of the child is known yet, set it.
      idxEnd = idxBegin + lengthElement;
      if(idxEnd > data.length)
        throw new AccessException("length current child is to large", this);
      if(parent != null && parent.currentChild == this)
      { //the typical case, it is the current child of something, 
        parent.idxCurrentChildEnd = idxEnd;  //next or addChild will use this start position for next currentChild.
        parent.correctIdxEnd(idxEnd);        //if the idxEnd of parents is less than new defined idxEnd of this
      }    
    }
    else if(idxEndP >0)
    { //the length of the child is not determinable yet, let it open to set later.
      idxEnd = idxEndP;  //the known size of whole data may be child of child. 
    }
    else
    { idxEnd = idxCurrentChildEnd; //it is after head.
    }
  }
  
  
  
  
  
  
  
  
  
  
  /**Older form, see protected method {@link assignCasted_i(ByteDataAccess, int )}
   * If a cast is possible, it should be programmed in the derivated class.
   * @deprecated because it is not necessary it's a downcast, it may be also upcast or sidecast. 
   * @param input
   * @throws AccessException
   */
  final protected void assignDowncast_i(ByteDataAccess input)
  throws AccessException
  { assignCasted_i(input, 0, -1);
  }
  
  
  

  /**assigns the element to the current child position of parent,
   * to represent the current child of the parent.
   * This method should be used by reading out byte[] data 
   * if it is detected, that the content of data matches of this derivated type of
   * ByteDataAccess. {@link next()} should be called before. The pattern of using is:
   * <pre>
   * 
    int nWhat;    //code of the element
    while( (nWhat = dataElement.next()) != ByteDataAccess.kEndOfElements )
    { switch(nWhat)  //test the first byte of the current element
      { case ByteDataAccess.kText:
        { sText = dataElement.getText();
        } break;
        case code1:
        { code1Element.assignAsChild(dataElement);
          evaluateValue(code1Element);
        } break:
        case code2:
        { code1Element.assignAsChild(dataElement);
          evaluateValue(code1Element);
        } break:
        default: throw new AccessException("unknown Element", dataElement);
      }
    }
    </pre>
   * The difference to {@link addChild(ByteDataAccess)} is: addChild() is used 
   * to writeout to data, addChild() appends the child always after idxEnd,
   * but this method is used to read from data and appends the child at position of the current child.
   * <br>
   * The data reference is copied, the idxBegin of this element
   * is set to the idxChild of parent, it is the current child position.
   * All other indices are setted calling {@link specifyLengthElementHead()}: idxChild
   * and {@link specifyLengthElement()}: idxEnd.
   * The idxChildEnd of parent is setted, so calling next() after this operation
   * increments in data after this new child.
   * <br>
   * If the element are using before, its connection to an other parent is dissolved.
   * @param parent The parent. It should reference data, and a current child position
   *        should be set by calling next() before. See sample at {@link next()}.
   * @throws AccessException If the data are wrong. The exception is thrown
   *        orginal from {@link specifyLengthElement()}.
   * @deprecated use addChild()
   */
  final public void assignAsChild(ByteDataAccess parent)
  throws AccessException
  { this.data = parent.data;
    this.bBigEndian = parent.bBigEndian;
    idxBegin = parent.idxCurrentChild;  //to mark start.
    idxCurrentChild = idxBegin + specifyLengthElementHead();
    idxEnd = idxBegin + specifyLengthElement();
    idxCurrentChildEnd = -1;
    parent.idxCurrentChildEnd = idxEnd;
    if(this.parent!= null && this.parent.currentChild == this) this.parent.currentChild = null;
    this.parent = null;  //the parent is not known, no add().
    currentChild = null;
    //idxEnd = xxx
    assignDataToFixChilds();
    //normally the idxEnd should dedicate the whole data area including data where no children are assigned.
    //but in some cases it doesn't so, than expand the data area. 
    if(parent.idxEnd < idxEnd)
    { //if it is added to empty data or the idxEnd is not set beforehand:
      parent.correctIdxEnd(idxEnd);  //the new end.
    }
  }




  
  /**assigns the element to the given position of the parents data to present a child of the parent.
   * The difference to {@link assignAsFromParentsChild(ByteDataAccess)} is: The position is given here
   * directly, it should not be the current child but a free child.  
   * <br>
   * The data reference is copied, the idxBegin of this element
   * is set to the idxChild given as parameter.
   * All other indices are set calling {@link specifyLengthElementHead()}: idxChild
   * and {@link specifyLengthElement()}: idxEnd.
   * <br>
   * If the element are using before, its connection to an other parent is dissolved.
   * @param parent The parent. It should reference data.
   * @param idxChildInParent The index in the data.
   * @throws AccessException If the data are wrong. The exception is thrown
   *        orginally from {@link specifyLengthElement()}.
   */
  final public void assignAtIndex(int idxChildInParent, ByteDataAccess parent)
  throws AccessException
  { this.data = parent.data;
    this.bBigEndian = parent.bBigEndian;
    if(this.parent!= null && this.parent.currentChild == this) this.parent.currentChild = null;
    this.parent = null;  //the parent is not known, no add().
    idxBegin = parent.idxBegin + idxChildInParent;  //to mark start.
    initIdxChildEndFromSpecify(parent.idxEnd);
    currentChild = null;
    //NOTE: childs help to assign Data if the parent-child-structure is built before assign().
    assignDataToFixChilds();
  }


  
  /**This method is called inside all assign...() methods. 
   * It should be overridden by a users derivation if any fix childs are present. 
   * In the overridden routines the method {@link assignDataAtIndex(byte[], int)} should be called for the fix childs.
   * example:<pre>
   * @Override protected void assignDataToFixChilds()
   * { myFixChild.assignDataAtIndex(super.data, super.idxBegin +kIdxmyFixChild);
   * } 
   * <pre>
   * Note that the indices are relative to idxBegin. use <code>super.</code> because an own element data or idxBegin may be present.
   * <br>
   * The originally implementation is empty.
   */
  protected void assignDataToFixChilds()
  {
  }


  /**adds an child Element after the current child or as first child after head.
   * With the aid of the child Element the data can be read or write structured. 
   *<br>
   * Some children can be added after a parent like following sample:
   * <pre>
   * ByteAccessDerivation child = new ByteAccessDerivation();  //empty and unassigned.
   * parent.addChild(child);        //The byte[] data of parent are assigned, index after current child index of parent.
   * child.addChild(grandchild);    //By adding a child to this child, also the parent's index is corrected.
   * </pre>
   *
   * @param child The child will be assigned with the data of this at index after the current child's end-index.
   * @throws AccessException if the length of the old current child is not determined yet.
   *         Either the method specifyLengthElement() should be overwritten or the method 
   *         {@link setLengthElement(int)} for the child or {@link setLengthCurrentChildElement(int)}
   *         should be called to prevent this exception.  
   * @throws AccessException if the length of the new current child is determined,
   *         but the length is to far for the data.
   */
  final public void addChild(ByteDataAccess child) 
  throws AccessException
  { notifyAddChild();

    child.data = data;
    child.bBigEndian = bBigEndian;
    child.parent = this;
    child.currentChild = null;
    this.currentChild = child;

    if(idxCurrentChildEnd > idxCurrentChild )
    { idxCurrentChild = idxCurrentChildEnd;
    }
    else if(idxCurrentChild ==-1)
    { //there is no child assigned yet, it is the first one.
      idxCurrentChild = idxBegin + specifyLengthElementHead();
    }
    else
      throw new AccessException("length of current child is undefined.", this); 
    child.idxBegin = idxCurrentChild;  //idxEnd; not idxEnd, if the data are written,idxChild should be the same as idxEnd. 
    child.initIdxChildEndFromSpecify(-1);
    child.assignDataToFixChilds();
    if(child.idxCurrentChild > idxEnd)
    { //the child is appended to empty data, expand the idxend
      child.correctIdxEnd(child.idxCurrentChild);  //the new end.
    }
  }

  
  
  /**Adds a child Element at current end of data to write data. 
   * The child's data are initialized with call of <code>child.specifyEmptyDefaultData().</code>
   *
   * @param child The child will associated to this and should be used
   *              to add some content.
   * @throws AccessException 
   */
  final public void addChildEmpty(ByteDataAccess child) 
  throws AccessException
  { addChild(child);
    child.specifyEmptyDefaultData();
  }
  
  

  /** Notifies, that a child is added. This method may be overload,
   * if the user must take somme actions, count the number of childs or others.
   */
  protected void notifyAddChild()
  { //in default, do nothing with this.
  }


  /** Increments the idxEnd and the idxCurrentChildEnd if a new child is added. Called only
   * inside method addChild(child) and recursively to correct
   * in all parents.
   */
  final protected void correctCurrentChildEnd(int idxEndNew)
  { if(idxEnd < idxEndNew) 
    { idxEnd = idxEndNew;
    }
    if(idxCurrentChildEnd < idxEndNew) 
    { idxCurrentChildEnd = idxEndNew;
    }
    if(parent != null)
    { parent.correctCurrentChildEnd(idxEndNew);
    }
  }

  /** Increments the idxEnd if a new child is added. Called only
   * inside method addChild(child) and recursively to correct
   * in all parents.
   */
  final protected void correctIdxEnd(int idxEndNew)
  { if(idxEnd < idxEndNew) 
    { idxEnd = idxEndNew;
    }
    if(parent != null)
    { parent.correctIdxEnd(idxEndNew);
    }
  }

  /** Expands the end index of the parent, it means the management
   * of the expanse of the data.
   *
   */
  final public void expandParent()
  throws AccessException
  { if(idxBegin == 0 && parent == null)
    { //it is the top level element, do nothing
    }
    else if(parent != null)
    { if(parent.idxEnd < idxEnd)
      { parent.idxEnd = idxEnd;
      }
      parent.expandParent();
    }
    else throw new AccessException("invalid expandParent()", this);
  }





  /** Writes a String into data with given color.
   * The user can overwrite this method and call super.addText(ss) inside
   * if some additional actions should be done.
   * @param ss The String to write
   * @param color The color
   */
  public int addText(String ss)
  throws AccessException
  { if(parent != null || idxBegin == 0)
    { byte[] chars = ss.getBytes();  //the ASCII represantation;
      int srcLen = chars.length;
      int dataLen = data.length;
      for(int ii=0; ii<srcLen; ii++)
      { byte cc = chars[ii];
        if(cc < 0x20) cc = 0x3f;  //'?' in ASCII
        if(idxEnd >= dataLen) throw new AccessException("element to long", this);
        data[idxEnd++] = cc;
      }
      expandParent();
      return(idxEnd-idxBegin);  //current nr of bytes
    }
    else throw new AccessException("inadmissable add-operation", this);
  }




  /** starts the calling loop of next().
   * The calling of next() after them supplies the first child element.
   *
   */
  final public void rewind()
  { idxCurrentChild = -1;
  }

  /** Sets the data index to the position after the current child element and returns its code.<br>
   * If more than 1 byte determines the code, the user should call getInt32(idxChild)
   * to get the code.
   * Usage:<pre>
   * while( (eElement = code.next()) != kEndOfElements)
   * { switch(eElement) { ... }
   * } </pre>
   * @return The first byte, that may be the code of the child element.<br>
   *         If this code is between 0x20 to 0xbf, it is assumed that is an
   *         character, kText is returned than.<br>
   *         If no more childs are available, kEndOfElements is returned.<br>
  */
  final public int next()
  throws AccessException
  { byte eWhat;

  
    if(idxCurrentChild ==-1)
    { //no child was assigned.
      idxCurrentChild = idxBegin + specifyLengthElementHead();
    }
    else if(idxCurrentChildEnd > idxCurrentChild )
    { //there is no child assigned yet, it is the first one.
      idxCurrentChild = idxCurrentChildEnd;
    }
    else
    { idxCurrentChild += getLengthCurrentChildElement();
    }
    idxCurrentChildEnd = -1;  //the child content is not checked, this index will be set if setLengthCurrentChildElement() is called.
  
    /*
    //set idxChild to the first or next element.
    if(idxChild < 0)
    {
      idxChild = idxBegin + specifyLengthElementHead();
      idxChildEnd = -1;
    }
    else
    { idxChild += getLengthCurrentChildElement();
    }
    */
    if( idxCurrentChild >= idxEnd)
    { eWhat = kEndOfElements;  //no more data
    }
    else
    { eWhat = data[idxCurrentChild];
      if(isTextByte(eWhat)){ eWhat = kText; }
    }
    return eWhat;
  }


  /**returns true if the given number of bytes is sufficing in the data from position of next child. 
   * 
   * @param nrofBytes that should fitting in the given data range from current child position 
   *                  to the end of data determines by calling assingData(...)
   *                  or by calling addChild() with a known size of child or setLengthElement() .
   * @return true if it is okay, false if the nrofBytes are negative or to large.
   * @throws AccessException see {@link getMaxNrofBytesForNextChild()} 
   */ 
  final public boolean sufficingBytesForNextChild(int nrofBytes) 
  throws AccessException
  { int maxNrofBytesChild = getMaxNrofBytesForNextChild();
    return nrofBytes < 0 ? false : maxNrofBytesChild >= nrofBytes;
  }
  
  
  /**returns the number number of bytes there are max available from position of a next current child. 
   * ,
   * the number of bytes to the end of buffer is returned.
   * 
   * @return nrofBytes that should fitting in the given data range from current child position 
   *                  to the end of data determines by calling assingData(...)
   *                  or by calling addChild() with a known size of child or setLengthElement() .
   * @throws AccessException if the length of the current child is not determined yet.
   *         Either the method specifyLengthElement() should be overwritten or the method 
   *         {@link setLengthElement(int)} for the child or {@link setLengthCurrentChildElement(int)}
   *         should be called to prevent this exception.  
   */ 
  final public int getMaxNrofBytesForNextChild() throws AccessException
  { //if(idxCurrentChild == -1)
    { //there is no child added, it is the first:
      //return idxEnd - idxCurrentChild;
    }  
    if(idxCurrentChildEnd <= idxCurrentChild)
      throw new AccessException("length of current child is undefined.", this); 
    return idxEnd - idxCurrentChildEnd;
  }
  
  
  /**returns the number number of bytes there are max available from position of the current child. 
   * ,
   * the number of bytes to the end of buffer is returned.
   * 
   * @return nrofBytes that should fitting in the given data range from current child position 
   *                  to the end of data determines by calling assingData(...)
   *                  or by calling addChild() with a known size of child or setLengthElement() .
   */ 
  final public int getMaxNrofBytes()
  { return data.length - idxBegin;
  }
  
  
  

  /** Returns the length of the existing actual element.
   * @return The number of bytes of the actual element in the buffer.
   *         It is (idxEnd - idxBegin).
   */
  final public int getLength()
  { return idxEnd - idxBegin;
  }



  /** Returns the length of the data.
   * @return The number of bytes of data in the buffer.
   *         It is idxEnd.
   */
  final public int getLengthTotal()
  { return idxEnd;
  } 



  /** Returns the data buffer itself. The actual total length is getted with getLengthTotal().
   * @return The number of bytes of the data in the buffer.
   */
  final public byte[] getData()
  { return data;
  }

  /**Returns the position of the Element data in the assigned buffer.
   * 
   * @return index of this element in the data buffer.
   */  
  final public int getPositionInBuffer()
  { return idxBegin;
  }
  


  /**Returns the position of the current child in the assigned buffer.
   * 
   * @return index of the current child of this element in the data buffer.
   */  
  final public int getPositionNextChildInBuffer()
  { return idxCurrentChildEnd;
  }
  



  /** Returns the length of the current child element.
   * The length may be setted outside by calling setLengthCurrentChildElement()
   * from user level after any calling of next()
   * or after calling getText() if it is a text content.
   * In this case this method returns this length in a simple way. <br>
   * If the user don't have called setLengthCurrentChildElement() after the last next(),
   * the users defined specifyLengthCurrentChildElement() is called.
   * @return the length in bytes of the current child element.
   * @throws AccessException if the user has not defined a overloaded methode specifyLengthCurrentChildElement()
   *                              or this method has thrown the exception because the length is not determinable.
   */
  final public int getLengthCurrentChildElement()
  throws AccessException
  {
    if(idxCurrentChildEnd >= 0)
    { //a get method, especially getText() was called,
      //so the end of the child is known yet, use it!
      return idxCurrentChildEnd - idxCurrentChild;
    }
    /*changed: it cannot be assumed that the coding use textbytes!
     * If they are textbytes, the user has called getText normally,
     * therefor the idxChildEnd is setted there.
    else if(isTextByte(data[idxChild]))
    { //text bytes are overreaded:
      idxChildEnd = idxChild;
      do
      { idxChildEnd +=1;
      } while(idxChildEnd < idxEnd && isTextByte(data[idxChildEnd]));
      return idxChildEnd - idxChild;
    }
    */
    else
    { //only the user can define the length.
      return specifyLengthCurrentChildElement();
    }
  }



  /** Sets the length of the current child element after calling next().
   *  The user may set the length due to the knowledge of the type and content of the actual child element.
   *  So the calling of specifyLengthCurrentChildElement() will be precluded.
   *  The idxChildEnd is setted.
   */
  final public void setLengthCurrentChildElement(int lengthOfCurrentChild)
  { if(currentChild != null)
    { currentChild.setLengthElement(lengthOfCurrentChild);
    }
  }


  
  /**Sets the length of the current element, considering all children.
   * If the element is the current child of its parent, the idxChildEnd of parent is set.
   * Therefore a call of next() or addChild uses the idxChildEnd-position for a new child.
   * 
   * @param length The length inclusive all children.
   */
  final public void setLengthElement(int length)
  { idxEnd = idxBegin + length;
    if(parent != null && parent.currentChild == this)
    { //the typical case, it is the current child of something, 
      parent.correctCurrentChildEnd(idxEnd);
    }
  }

  
  
  
  
  
  /** Returns true if the current child element represents a TEXT(), direct ASCII chars,
   * false if the element is a complex element.
   * Text chararacters are coded with 0x20..0x7f, the standard ASCII code,
   * and with 0x80 to 0xbf, special characters user defined.
   *
   */
  public boolean isTextByte(byte nn)
  { return (nn < -0x40 || nn >= 0x20);
  }


  /** Returns the current string or null on end*/
  final public String getText()
  { if(idxCurrentChild >= idxEnd) return null;
    else
    { byte nn = data[idxCurrentChild];
      if( isTextByte(nn))
      { idxCurrentChildEnd = idxCurrentChild;
        char cc = '?';
        StringBuffer ss = new StringBuffer(20);
        while(  idxCurrentChildEnd < idxEnd
             && isTextByte(nn = data[idxCurrentChildEnd])
             )
        { if(nn < 0) cc = '?'; //{ cc=convertHiChars(nn);}  //range between 0x80..0xbf
          else cc = (char)(nn);     //convert 0x20...7F to character
          ss.append(cc);  //konvert to Unicode
          idxCurrentChildEnd +=1;
        }
        return ss.toString();
      }
      else return null;  //no Text
    }
  }

  /**Returns a String from the given position inside the actual element .
   * The bytes are interpreted in the given encoding.
   * 
   * @param idx The start position inside the child.
   * @param nmax Maximal number of bytes
   * @return The String representation of the bytes.
   */
  protected String getString(int idx, int nmax)
  { String sRet;
    try{ sRet = new String(data, idxBegin + idx, nmax, "ISO-8859-1");} 
    catch (UnsupportedEncodingException e)  {sRet = null; }
    int pos0 = sRet.indexOf(0);
    if(pos0 >0 )
    { //The data are zero terminated!
      sRet = sRet.substring(0, pos0);
    }
    return sRet;
  }
  
  

  /**Sets a String to the the given position inside the actual element .
   * The bytes are interpreted in the given encoding.
   * 
   * @param idx The start position inside the child.
   * @param nmax Maximal number of bytes
   * @param ss The String representation of the bytes.
   * @return Number of written chars.
   * @throws  
   */
  protected int setString(int idx, int nmax, String ss)
  { if(ss.length()>nmax){ ss = ss.substring(0, nmax); } //truncate.
    byte[] bytes;
    try { bytes = ss.getBytes("ISO-8859-1");} 
    catch (UnsupportedEncodingException e){ bytes = null; }
    int len = bytes.length;
    if(len > nmax){ len = nmax; } //truncate.
    System.arraycopy(bytes, 0, data, idxBegin + idx, len);
    return len;
  }
  
  

  /** Sets the big or little endian mode.
   *
   * @param val true if big endian, hi byte at lower adress, false if little endian.
   */
  protected void setBigEndian(boolean val)
  { bBigEndian = val;
  }

  /** Returns the content of 1 to 8 bytes inside the actual element as a long number,
   * big- or little-endian depending on setBigEndian().
   * This method is protected because at user level its using is a prone to errors because the idx is free related.
   *
   * @param idx the position of leading byte in the actual element, the data are taken from data[idxBegin+idx].
   * @return the long value in range adequate nrof bytes.
   * */
  protected final long _getLong(int idx, final int nrofBytes)
  { int val = 0;
    int idxStep;
    if(bBigEndian)
    { idx = idxBegin + idx;
      idxStep = 1;
    }
    else
    { idx = idxBegin + idx + nrofBytes -1;
      idxStep = -1;
    }
    int nByteCnt = nrofBytes;
    do
    { val |= data[idx] & 0xff;
      if(--nByteCnt <= 0) break;  //TRICKY: break in mid of loop, no shift operation.
      val <<=8;
      idx += idxStep;
    }while(true);  //see break;
    int posSign = (nrofBytes*8)-1;  //position of sign of the appropriate nrofBytes 
    if( (val & (1<<posSign)) != 0)
    { val |= 0xffffffffffffffffL << (posSign+1);  //supplement the rest bits of long with the sign value,it's negativ.   
    }
    return val;
  }

  
  /** Returns the content of 1 to 8 bytes inside the actual element as a long number,
   * big- or little-endian depending on setBigEndian().
   * This method is protected because at user level its using is a prone to errors because the idx is free related.
   *
   * @param idx the position of leading byte in the actual element, the data are taken from data[idxBegin+idx].
   * @return the long value in range adequate nrof bytes.
   * */
  protected final void _setLong(int idx, int nrofBytes, long val)
  { int idxStep;
    if(bBigEndian)
    { idx = idxBegin + idx + nrofBytes -1;
      idxStep = -1;
    }
    else
    { idx = idxBegin + idx;
      idxStep = 1;
    }
    do
    { data[idx] = (byte)(val);
      if(--nrofBytes <= 0) break;
      val >>=8;
      idx += idxStep;
    }while(true);  //see break;
  }

  
  /**copies some data to a int[], primarily to debug a content. 
   * @param dst This array is field, but only from data of the current element between idxBegin and idxEnd
   */
  public final void copyData(int[] dst)
  { int idxMax = idxEnd - idxBegin;
    if(idxMax/4 > dst.length) idxMax = 4* dst.length;
    int iDst = 0;
    for(int idx = 0; idx < idxMax; idx+=4)
    { dst[iDst++] = (int)_getLong(idx,4);
    }
  }
  
  
  
  
  /** Gets a float value from the content of 4 byte. The float value is red
   * according to the IEEE 754 floating-point "single format" bit layout, preserving Not-a-Number (NaN) values,
   * like converted from java.lang.Float.intBitsToFloat().
   */
  protected final float getFloat(int idx)
  {
    int intRepresentation = getInt32(idx);
    return Float.intBitsToFloat(intRepresentation);
  }
  
  protected final double getDouble(int idx)
  {
    long intRepresentation = getInt64(idx);
    return Double.longBitsToDouble(intRepresentation);
   
  }
  
  protected final long getInt64(int idx)
  { int nLo,nHi;
    if(bBigEndian)
    { nLo = getInt32(idx);
      nHi =  getInt32(idx + 4);
    }
    else
    { nLo = getInt32(idx+4);
      nHi =  getInt32(idx);
    }
    long val = nHi << 32;
    val += nLo;
    return val;
  }

  /** Returns the content of 4 bytes inside the actual element as a integer number between -2147483648 and 2147483647,
   * big- or little-endian depending on setBigEndian().
   *
   * @param idx the position of leading byte in the actual element, the data are raken from data[idxBegin+idx].
   *            This is not the absolute position in data, idxBegin is added.<br/>
   * @return the integer value in range from -2147483648 and 2147483647
   * */
  public final int getInt32(int idx)
  { int val;
    if(bBigEndian)
    { val =  ((  (int)data[idxBegin + idx])<<24)  //NOTE all 24 low-bits are 0
          |  (( ((int)data[idxBegin + idx+1])<<16) & 0x00ff0000 ) //NOTE the high bits may be 0 or 1
          |  (( ((int)data[idxBegin + idx+2])<< 8) & 0x0000ff00 ) //depending on sign of byte. Mask it!
          |  (( ((int)data[idxBegin + idx+3])    ) & 0x000000ff );  //NOTE: the value has only 8 bits for bitwise or.
    }
    else
    { val =  (  ((int)data[idxBegin + idx+3])<<24)  //NOTE all 24 low-bits are 0
          |  (( ((int)data[idxBegin + idx+2])<<16) & 0x00ff0000 ) //NOTE the high bits may be 0 or 1
          |  (( ((int)data[idxBegin + idx+1])<< 8) & 0x0000ff00 ) //depending on sign of byte. Mask it!
          |  (( ((int)data[idxBegin + idx  ])    ) & 0x000000ff );  //NOTE: the value has only 8 bits for bitwise or.
    }
    return val;
  }

  
  public final int getUint32(int idx)
  { return getInt32(idx);
  }
  
  /** Returns the content of 2 bytes as a positive nr between 0..65535, big-endian
   * inside the actual element.
   *
   * @param idx the position of leading byte in the actual element, the data are raken from data[idxBegin+idx].
   *            This is not the absolute position in data, idxBegin is added.<br/>
   * @return the integer value in range from -32768 to 32767
   * */
  protected final int getUint16(int idx)
  { int val;
    if(bBigEndian)
    { val =  (( ((int)data[idxBegin + idx  ])<< 8) & 0x0000ff00 ) //depending on sign of byte. Mask it!
          |  (( ((int)data[idxBegin + idx+1])    ) & 0x000000ff );  //NOTE: the value has only 8 bits for bitwise or.
    }
    else
    { val =  (( ((int)data[idxBegin + idx+1])<< 8) & 0x0000ff00 ) //depending on sign of byte. Mask it!
          |  (( ((int)data[idxBegin + idx  ])    ) & 0x000000ff );  //NOTE: the value has only 8 bits for bitwise or.
    }
    return val;
  }

  /** Returns the content of 2 bytes as a positive nr between 0..65535, big-endian
   * inside the actual element.
   *
   * @param idx the position of leading byte in the actual element, the data are raken from data[idxBegin+idx].
   *            This is not the absolute position in data, idxBegin is added.<br/>
   * @return the integer value in range from -32768 to 32767
   * */
  protected final int getInt16(int idx)
  { int val;
    if(bBigEndian)
    { val =  (( ((int)data[idxBegin + idx  ])<< 8) & 0xffffff00 ) //depending on sign of byte. Mask it!
          |  (( ((int)data[idxBegin + idx+1])    ) & 0x000000ff );  //NOTE: the value has only 8 bits for bitwise or.
    }
    else
    { val =  (( ((int)data[idxBegin + idx+1])<< 8) & 0x0000ff00 ) //depending on sign of byte. Mask it!
          |  (( ((int)data[idxBegin + idx  ])    ) & 0x000000ff );  //NOTE: the value has only 8 bits for bitwise or.
    }
    return val;
  }

  /** Returns the content of 1 bytes as ASCII
   * inside the actual element.
   *
   * @param idx the position of char in the actual element, the data are raken from data[idxBegin+idx].
   *            This is not the absolute position in data, idxBegin is added.<br/>
   * @return the ASCII value
   * */
  protected final char getChar(int idx)
  { char val;
    val = (char) data[idxBegin + idx];
    return val;
  }
  
  /** Returns the content of 1 bytes as a positive or negative nr between -128..127
   * inside the actual element.
   *
   * @param idx the position of leading byte in the actual element, the data are raken from data[idxBegin+idx].
   *            This is not the absolute position in data, idxBegin is added.<br/>
   * @return the integer value in range from -32768 to 32767
   * */
  protected final int getInt8(int idx)
  { int val;
    val = data[idxBegin + idx];
    return val;
  }

  
  /** Returns the content of 1 bytes as a positive or negative nr between -128..127
   * inside the actual element.
   *
   * @param idx the position of leading byte in the actual element, the data are raken from data[idxBegin+idx].
   *            This is not the absolute position in data, idxBegin is added.<br/>
   * @return the integer value in range from -32768 to 32767
   * */
  protected final int getUint8(int idx)
  { int val;
    val = data[idxBegin + idx] & 0xff;
    return val;
  }
 
  
  
  

  
  protected final int getUint32(int idxBytes, int idxArray, int lengthArray)
  { if(idxArray >= lengthArray || idxArray < 0) throw new IndexOutOfBoundsException("getUint16:" + idxArray);
    return getUint32(idxBytes + 4*idxArray);
  }
  
  protected final int getInt32(int idxBytes, int idxArray, int lengthArray)
  { if(idxArray >= lengthArray || idxArray < 0) throw new IndexOutOfBoundsException("getInt32:" + idxArray);
    return getInt32(idxBytes + 4*idxArray);
  }
  
  protected final int getInt16(int idxBytes, int idxArray, int lengthArray)
  { if(idxArray >= lengthArray || idxArray < 0) throw new IndexOutOfBoundsException("getInt16:" + idxArray);
    return getInt16(idxBytes + 2*idxArray);
  }
  
  protected final int getInt8(int idxBytes, int idxArray, int lengthArray)
  { if(idxArray >= lengthArray || idxArray < 0) throw new IndexOutOfBoundsException("getInt8:" + idxArray);
    return getInt8(idxBytes + idxArray);
  }
  
  protected int getUint16(int idxBytes, int idxArray, int lengthArray)
  { if(idxArray >= lengthArray || idxArray < 0) throw new IndexOutOfBoundsException("getUint16:" + idxArray);
    return getUint16(idxBytes + 2*idxArray);
  }
  
  protected int getUint8(int idxBytes, int idxArray, int lengthArray)
  { if(idxArray >= lengthArray || idxArray < 0) throw new IndexOutOfBoundsException("getUint8:" + idxArray);
    return getInt8(idxBytes + idxArray);
  }
  
  protected float getFloat(int idxBytes, int idxArray, int lengthArray)
  { if(idxArray >= lengthArray || idxArray < 0) throw new IndexOutOfBoundsException("getFloat:" + idxArray);
    return getFloat(idxBytes + 4*idxArray);
  }
  
  
  
  

  /** Set the content of 4 byte from a float variable. The float value is stored
   * according to the IEEE 754 floating-point "single format" bit layout, preserving Not-a-Number (NaN) values,
   * like converted from java.lang.Float.floatToRawIntBits().
   */
  protected void setFloat(int idx, float value)
  {
    int intRepresentation = Float.floatToRawIntBits(value);
    setInt32(idx, intRepresentation);
  }



  /** Set the content of 4 bytes as a integer number between -2147483648 and 2147483647,
   * big- or little-endian depended from setBigEndian().
   *
   * @param idx The position of leading byte in the current elements data.
   *            This is not the absolute position in data, idxBegin is added.<br/>
   * @param value The value in range 0..65535. The value is taken modulo 0xffffffff.
   * */
  protected void setInt32(int idx, int value)
  { if(bBigEndian)
    { data[idxBegin + idx]   = (byte)((value>>24) & 0xff);
      data[idxBegin + idx+1] = (byte)((value>>16) & 0xff);
      data[idxBegin + idx+2] = (byte)((value>>8) & 0xff);
      data[idxBegin + idx+3] = (byte)(value & 0xff);
    }
    else
    { data[idxBegin + idx+3] = (byte)((value>>24) & 0xff);
      data[idxBegin + idx+2] = (byte)((value>>16) & 0xff);
      data[idxBegin + idx+1] = (byte)((value>>8) & 0xff);
      data[idxBegin + idx]   = (byte)(value & 0xff);
    }
  }

  /** Set the content of 1 bytes as a positive nr between 0..255, big- or little-endian.
  *
  * @param idx The position of leading byte in the current elements data.
  *            This is not the absolute position in data, idxBegin is added.<br/>
  * @param value The value in range 0..65535. The value is taken modulo 0xff.
  * */
  protected void setUint8(int idx, int value)
  { setInt8(idx, value);  //its the same because modulo!
  }

  /** Set the content of 2 bytes as a positive nr between 0..65535, big- or little-endian.
  *
  * @param idx The position of leading byte in the current elements data.
  *            This is not the absolute position in data, idxBegin is added.<br/>
  * @param value The value in range 0..65535. The value is taken modulo 0xffff.
  * */
  protected void setUint16(int idx, int value)
  { setInt16(idx, value);  //its the same because modulo!
  }

  /** Set the content of 4 bytes as a positive nr between 0..2pow32-1, big- or little-endian.
  *
  * @param idx The position of leading byte in the current elements data.
  *            This is not the absolute position in data, idxBegin is added.<br/>
  * @param value The value as long. The value is taken modulo 0xffffffff.
  * */
  protected void setUint32(int idx, long value)
  { //the same algorithm in source, but other action on machine level,
    //because value is long!
    if(bBigEndian)
    { data[idxBegin + idx]   = (byte)((value>>24) & 0xff);
      data[idxBegin + idx+1] = (byte)((value>>16) & 0xff);
      data[idxBegin + idx+2] = (byte)((value>>8) & 0xff);
      data[idxBegin + idx+3] = (byte)(value & 0xff);
    }
    else
    { data[idxBegin + idx+3] = (byte)((value>>24) & 0xff);
      data[idxBegin + idx+2] = (byte)((value>>16) & 0xff);
      data[idxBegin + idx+1] = (byte)((value>>8) & 0xff);
      data[idxBegin + idx]   = (byte)(value & 0xff);
    }
  }

  /** Set the content of 2 bytes from an integer between -32768..32768,
   * or from an integer number between 0..65535. The value is interpreted
   * from the input parameter with modulo 0x10000.
   * Big- or little-endian.
   *
   * @param idx The position of leading byte in the current elements data.
   *            This is not the absolute position in data, idxBegin is added.<br/>
   * @param value The value in range 0..65535. The value is taken modulo 0xffff.
   * */
  protected void setInt16(int idx, int value)
  { if(bBigEndian)
    { data[idxBegin + idx]   = (byte)((value>>8) & 0xff);
      data[idxBegin + idx+1] = (byte)(value & 0xff);
    }
    else
    { data[idxBegin + idx+1] = (byte)((value>>8) & 0xff);
      data[idxBegin + idx]   = (byte)(value & 0xff);
    }
  }

  /** Set the content of 1 bytes as a positive nr between 0..256.
   *
   * @param idx The position of leading byte in the current elements data.
   *            This is not the absolute position in data, idxBegin is added.<br/>
   * @param value The value in range 0..65535. The value is taken modulo 0xffff.
   * */
  protected void setInt8(int idx, int value)
  { data[idxBegin + idx] = (byte)(value & 0xff);
  }


  
  protected void setUint32(int idxBytes, int idxArray, int lengthArray, int val)
  { if(idxArray >= lengthArray || idxArray < 0) throw new IndexOutOfBoundsException("setUint32:" + idxArray);
    setUint32(idxBytes + 4*idxArray, val);
  }
  
  protected void setInt32(int idxBytes, int idxArray, int lengthArray, int val)
  { if(idxArray >= lengthArray || idxArray < 0) throw new IndexOutOfBoundsException("setInt32:" + idxArray);
    setInt32(idxBytes + 4*idxArray, val);
  }
  
  protected void setInt16(int idxBytes, int idxArray, int lengthArray, int val)
  { if(idxArray >= lengthArray || idxArray < 0) throw new IndexOutOfBoundsException("getInt16:" + idxArray);
    setInt16(idxBytes + 2*idxArray, val);
  }
  
  protected void setInt8(int idxBytes, int idxArray, int lengthArray, int val)
  { if(idxArray >= lengthArray || idxArray < 0) throw new IndexOutOfBoundsException("getInt16:" + idxArray);
    setInt8(idxBytes + idxArray, val);
  }
  
  
  protected void setUint16(int idxBytes, int idxArray, int lengthArray, int val)
  { if(idxArray >= lengthArray || idxArray < 0) throw new IndexOutOfBoundsException("getInt16:" + idxArray);
    setUint16(idxBytes + 2*idxArray, val);
  }
  
  protected void setUint8(int idxBytes, int idxArray, int lengthArray, int val)
  { if(idxArray >= lengthArray || idxArray < 0) throw new IndexOutOfBoundsException("getInt16:" + idxArray);
    setUint8(idxBytes + idxArray, val);
  }
  
  protected void setFloat(int idxBytes, int idxArray, int lengthArray, float val)
  { if(idxArray >= lengthArray || idxArray < 0) throw new IndexOutOfBoundsException("getInt16:" + idxArray);
    setFloat(idxBytes + 4 * idxArray, val);
  }
  
  
  
  
  /** Copies the data into a byte[]
   *
   * */
  final public void copyDataFrom(ByteDataAccess src)
  throws AccessException
  { int len = src.getLength();
    if(data.length < len) throw new IndexOutOfBoundsException("copy, dst to small" + len);
    System.arraycopy(src.data,src.idxBegin,data,idxBegin,len);
  }

/** Counts the idxChild by given index, idxChild is ByteCount from idxBegin
 * 
 * @param indexObjectArray Index of Array
 */
  public void elementAt(int indexObjectArray) {
      
      try {
          idxCurrentChild = idxBegin + specifyLengthElementHead() + specifyLengthCurrentChildElement() * indexObjectArray;
      } catch (AccessException e) {
          e.printStackTrace();
      }
      
  }



  public ByteDataAccess getCurrentChild() 
  {
     
      return currentChild;
  }

  
  
  public class AccessException extends Exception
  {
    /** The causer of the Exception, the idx... are important.
     */
    ByteDataAccess causer;
    /**
     *
     */
    private static final long serialVersionUID = 7853831657510575331L;
  
    public AccessException()
    {
      super();
      causer = null;
    }
  
    public AccessException(String message, ByteDataAccess causer)
    { super(message);
      this.causer = causer;
    }
  
    public AccessException(Throwable cause)
    {
      super(cause);
    }
  
    public AccessException(String message, Throwable cause)
    {
      super(message, cause);
    }
  
  
  }
  
  
  
  
}
