/****************************************************************************
 * Copyright/Copyleft:
 *
 * For this source the LGPL Lesser General Public License,
 * published by the Free Software Foundation is valid.
 * It means:
 * 1) You can use this source without any restriction for any desired purpose.
 * 2) You can redistribute copies of this source to everybody.
 * 3) Every user of this source, also the user of redistribute copies
 *    with or without payment, must accept this license for further using.
 * 4) But the LPGL ist not appropriate for a whole software product,
 *    if this source is only a part of them. It means, the user
 *    must publish this part of source,
 *    but don't need to publish the whole source of the own product.
 * 5) You can study and modify (improve) this source
 *    for own using or for redistribution, but you have to license the
 *    modified sources likewise under this LGPL Lesser General Public License.
 *    You mustn't delete this Copyright/Copyleft inscription in this source file.
 *
 * @author JcHartmut: hartmut.schorrig@vishia.de
 * @version 2008-04-06  (year-month-day)
 * list of changes:
 * 2008-04-06 JcHartmut: some correction
 * 2008-03-15 JcHartmut: creation
 *
 ****************************************************************************/
package org.vishia.java2C;

import java.util.LinkedList;
import java.util.List;
import java.util.TreeMap;

import org.vishia.zbnf.ZbnfParseResultItem;


/**This class represents the content of a java class for Translation.
 * The instance is created and filled while executing the first pass of translation, the generation of header.
 * It is used for the second pass, the generation of C-file.
 * <br>
 * Chars are used to identify the kind of a element or type:
 * <table>
 * <tr><td>$</td><td>an embedded struct</td></tr>
 * <tr><td>%</td><td>a variable of basic scalar type such as int32, float, char</td></tr>
 * <tr><td>*</td><td>a normal reference, pointer in C, used for non-GC-controlled associations.</td></tr>
 * <tr><td>@</td><td>an enhanced reference, for GC-associations.</td></tr>
 * <tr><td>&</td><td>an interface enhanced reference, using C-jump-table with index, GC is possible</td></tr>
 * <tr><td>d</td><td>simple constant implemented with #define</td></tr>
 * <tr><td>s</td><td>a static field (element) of basic type</td></tr>
 * <tr><td>S</td><td>a final static field (element) of basic type with initializer, const in C</td></tr>
 * <tr><th></th><th>Types</th></tr>
 * <tr><td>C</td><td>a class type identifier</td></tr>
 * <tr><td>I</td><td>an interface type identifier</td></tr>
 * <tr><td>B</td><td>a basic scalar type</td></tr>
 * <tr><td>x</td><td>...</td></tr>
 * <tr><td>x</td><td>...</td></tr>
 * <tr><td>x</td><td>...</td></tr>
 * <tr><td>x</td><td>...</td></tr>
 * </table>
 */
class ClassData
{
  /**The full name, like produced in C. If it is an inner class, outer__inner. */
  public final String sName;
  
  
  /**The file were the class is defined in. Without extension .java.
   * 
   */
  public final String sFileName;
  
  /**Set if @staticInstance is found in description of the class.
   * A class is a static instance if the class should not be created dynamically using garbage collection,
   * but only in the construction path of the application.
   * In such classes all final references to final classes (may not be staticInstance itself)
   * are simple pointers in the C Conversion. Otherwise enhanced references are to be used.
   */
  public final boolean isStaticInstanciated;

  /**Set if the class is final. */
  public final boolean isFinal;

  private boolean bReferenceNeeded = false;

  /**For second pass, the parse result. NOTE: Not final because it should be removed after running second path. */
  private ZbnfParseResultItem itemClass;

  
  /**List of known types.*/
  public final TreeMap<String, String> typeIdentifiers = new TreeMap<String, String>();

  /**All known identifiers of this class. They are also fields of this class, 
   * types (= inner classes) of this class
   */
  public final LocalIdents classLevelIdents;
  
  
  /**The class as type info useable for references. */
  public final LocalIdents.FieldTypeInfos classTypeInfo;
  
  /**The superclass of this or null. */
  ClassData superClazz;
  
  
  /**The outer class of this or null. */
  ClassData outerClazz;
  
  /**This class saves the position of itemValue in a parse result to the named variable.
   * It is used if the initialization is generated.
   */
  static class VariablesToInit
  { /**The zbnf parse result element which contains the value. */
    final ZbnfParseResultItem itemValue;
    /**Name of the variable. */
    final String sName;
    /**Type of the variable. */
    final String sType;
    /**Kind of variable, use $:embedded s:static etc, */
    final char kind;
    
    VariablesToInit(char kind, String sType, String sName, ZbnfParseResultItem itemValue)
    { this.kind = kind; this.itemValue= itemValue; this.sName = sName; this.sType = sType; }
  }

  /**List of parse result items of initial values for variables. 
   * Filled in first path, used in second pass.
   */
  final List<VariablesToInit> variablesToInit = new LinkedList<VariablesToInit>();

  /**List of all static variables either with initial values or not.
   * Filled in first path, used in second pass.
   */
  final List<VariablesToInit> staticVariables = new LinkedList<VariablesToInit>();

  ClassData(String sFileName, String sName, ZbnfParseResultItem itemClass, boolean bIsStaticInstance, boolean bFinal)
  { this.sFileName = sFileName;
    this.sName = sName; this.itemClass= itemClass;
    this.isStaticInstanciated = bIsStaticInstance;
    this.isFinal = bFinal;
    this.classLevelIdents = new LocalIdents();
    classTypeInfo = new LocalIdents.FieldTypeInfos(sName, this, "C", this );
  }

  public void referenceNeeded(){ bReferenceNeeded = true; }

  public boolean isReferenceNeeded(){ return bReferenceNeeded; }

  String getName(){ return sName; }

  ZbnfParseResultItem getParseResult(){ return itemClass; }

  /**relinquishes the parse result, because it is processed. It should not consumed memory.*/
  void relinquishParseResult(){ itemClass = null; }

  
  //List<VariablesToInit> getVariablesToInit(){ return variablesToInit; }

  public String toString(){ return sName; }
  
}
