/****************************************************************************
 * Copyright/Copyleft:
 *
 * For this source the LGPL Lesser General Public License,
 * published by the Free Software Foundation is valid.
 * It means:
 * 1) You can use this source without any restriction for any desired purpose.
 * 2) You can redistribute copies of this source to everybody.
 * 3) Every user of this source, also the user of redistribute copies
 *    with or without payment, must accept this license for further using.
 * 4) But the LPGL ist not appropriate for a whole software product,
 *    if this source is only a part of them. It means, the user
 *    must publish this part of source,
 *    but don't need to publish the whole source of the own product.
 * 5) You can study and modify (improve) this source
 *    for own using or for redistribution, but you have to license the
 *    modified sources likewise under this LGPL Lesser General Public License.
 *    You mustn't delete this Copyright/Copyleft inscription in this source file.
 *
 * @author JcHartmut: hartmut.schorrig@vishia.de
 * @version 0.82
 * list of changes:
 * 2008-04-06 JcHartmut: some correction
 * 2008-03-15 JcHartmut: creation
 *
 ****************************************************************************/
package org.vishia.java2C;

import java.io.IOException;
import java.text.ParseException;
import java.util.Iterator;
import java.util.List;
import java.util.TreeMap;

import org.vishia.zbnf.ZbnfParseResultItem;

/**This class processes the first pass of translating Java to C of a detect class in a java file.
 * Parsing the content of the java file is done before, see {@link GenerateFile}.
 * In the first pass all declared elements of a class are captured and store in the {@link ClassData}.  
 * 
 * @author JcHartmut
 *
 */
public class FirstPass extends GenerateClass
{

  /**initializes the instances with the given associations. */ 
  FirstPass(iWriteContent writeContent, TreeMap<String, ClassData> allClassData, RunRequiredFirstPass_ifc runRequiredFirstPass)
  { super(writeContent, allClassData, runRequiredFirstPass);
  }


  /**Generates the content of a class in the h-file with given ZBNF parse result item of the class.
   * This routine is called from {@link GenerateFile}.  
   * <ul>
   * <li> All internal classes and interfaces are detect, this method is called recursively with it.
   * <li>The class fields are written in a struct in the h-File. 
   * <li> The static fields are written in h-file as extern declaration.
   * <li> The constructors are written as forward declaration in h-file of routines ctor...
   * <li> The methods are written as forward declaration in h-file.
   * </ul> 
   * 
   * @param sFileName Name of the h- and c-file. It is stored in the created ClassData of the class
   *        to allow the generating of a include statement, if another class uses this class.
   * @param itemClass The zbnf parser result item of the class.
   * @param sOuterClassName "" or the name of the outer class. 
   *        The name of the outer class is setted as prefix of the name of the struct in C 
   * @param itemsClassesForSecondPath List to add the detect ClassData. 
   *        The list is used as input for running the second path of classes. 
   *        See {@link GenerateFile.itemsClassesForSecondPath}.  
   * @return
   * @throws IOException
   * @throws ParseException
   */
  ClassData runFirstPass
  ( String sFileName
  , ZbnfParseResultItem itemClass
  , String sOuterClassName
  , List<ClassData> itemsClassesForSecondPath
  )
  throws IOException, ParseException
  {

    ZbnfParseResultItem itemClassDescription = itemClass.getChild("description");
    boolean bStaticInstance =  itemClassDescription != null
                            && itemClassDescription.getChild("staticInstance")!= null;

    ZbnfParseResultItem itemClassIdent = itemClass.getChild("classident");
    if(itemClassIdent == null)
      throw new ParseException("classident not detect", 0);
    String sClassName = sOuterClassName + itemClassIdent.getParsedString();

    { /**First process subclasses because there may be known before the outer class is translated. */
      List<ZbnfParseResultItem> listSubClasses = itemClass.listChildren("classDefinition");
      if(listSubClasses != null)
      { for(ZbnfParseResultItem itemSubClasses : listSubClasses)
        { FirstPass genClass = new FirstPass(writeContent, allClassData, runRequiredFirstPass);
          genClass.runFirstPass(sFileName, itemSubClasses, sClassName + "__", itemsClassesForSecondPath);
      } }
    }
    { List<ZbnfParseResultItem> listSubInterfaces = itemClass.listChildren("interfaceDefinition");
      if(listSubInterfaces != null)
      { for(ZbnfParseResultItem itemSubInterfaces : listSubInterfaces)
        { write_Interface(itemSubInterfaces, sClassName + "__");
      } }
    }

    boolean bFinal = itemClass.getChild("final")!= null;
    classData = new ClassData(sFileName, sClassName, itemClass, bStaticInstance, bFinal);
    allClassData.put(sClassName, classData);   //to find it later sorted by names.
    itemsClassesForSecondPath.add(classData);  //to process in the second path sorted per file

    setInitialTypeIdentifiers();

    writeContent.writeClassH("typedef struct " + sClassName + "_t\n{ ");
    writeContent.writeClassH("ObjectJc object;");

    if(sClassName == null) throw new ParseException("no className found", 0);

    //process the class Content, write all class variable in the struct
    List<ZbnfParseResultItem> listVariables = itemClass.listChildren("variableDefinition");
    if(listVariables != null)for(ZbnfParseResultItem itemVariable : listVariables)
    { if(itemVariable.getChild("static")== null)
      { //only non static members.
        String sDef = gen_variableDefinition(itemVariable, classData.classLevelIdents, classData.variablesToInit, 'c');
        writeContent.writeClassH("\n  " + sDef);
      }  
    }
    
    writeContent.writeClassH("\n} " + sClassName + ";\n  \n");

    if(listVariables != null)for(ZbnfParseResultItem itemVariable : listVariables)
    { if(itemVariable.getChild("static")!= null)
      { //only non static members.
        String sDef = gen_variableDefinition(itemVariable, classData.classLevelIdents, classData.variablesToInit, 's');
        writeContent.writeClassH("\n  " + sDef);
      }  
    }
    writeContent.writeClassH("\n\n");
    
    { List<ZbnfParseResultItem> listConstructors = itemClass.listChildren("constructorDefinition");
      if(listConstructors != null)
      { for(ZbnfParseResultItem itemConstructor : listConstructors)
        { write_constructorDeclaration(itemConstructor, sClassName);
        }
      }
      else
      { //no constructor found, create a default constructor because all variable initializations should be done.
        write_constructorDeclaration(null, sClassName);
      }
    }
    { List<ZbnfParseResultItem> listMethods = itemClass.listChildren("methodDefinition");
      if(listMethods != null)
      { for(ZbnfParseResultItem itemMethod : listMethods)
        { write_methodDeclaration(itemMethod, sClassName);
      } }
    }

    return classData;
  }



  /**Generates the content of a class in c- and h-file. */
  void write_Interface(ZbnfParseResultItem itemClass, String sOuterClassName)
  throws IOException, ParseException
  {
    //ZbnfParseResultItem resultItem = resultItemClass.nextSkipIntoComponent(resultItemClass);

    ZbnfParseResultItem itemClassIdent = itemClass.getChild("interfaceIdent");
    if(itemClassIdent == null) throw new ParseException("interfaceIdent not detect", 0);

    ZbnfParseResultItem itemDescription = itemClass.getChild("description");
    if(itemDescription != null)
    { write_Description(itemDescription);
    }


    String sClassName = sOuterClassName + itemClassIdent.getParsedString();
    writeContent.writeClassH("typedef struct " + sClassName + "_t  /*interface*/\n{");
    writeContent.writeClassH("ObjectJc object;");

    if(sClassName == null) throw new ParseException("no className found", 0);

    Iterator<ZbnfParseResultItem> iterMethod = itemClass.iterChildren("methodDefinition");
    if(iterMethod != null)
    { while(iterMethod.hasNext())
      { ZbnfParseResultItem itemMethod = iterMethod.next();
        write_methodDeclaration(itemMethod, sClassName);
    } }

    writeContent.writeClassH("\n} " + sClassName + ";\n  \n");


    List<ZbnfParseResultItem> listSubClasses = itemClass.listChildren("classDefinition");
    if(listSubClasses != null)
    { for(ZbnfParseResultItem itemSubClasses : listSubClasses)
      { //runFirstPass(itemSubClasses, sClassName + "__", itemsxxx);
      }
    }
  }



  /**writes a methodDefinition with syntax methodDefinition::=.
   *
   * @param parent
   * @throws IOException
   * @throws ParseException
   */
  private void write_constructorDeclaration(ZbnfParseResultItem parent, String sClassName)
  throws IOException, ParseException
  {
    String sDescription = parent!=null ? get_description(parent) : "Default constructor. ";
    if(sDescription != null)
    { writeContent.writeClassH("/**" + sDescription + "*/\n");
    }

    String sMethodHead = sClassName + "* ctor_" + sClassName + "(MemC rawMem";
    if(parent!=null)
    { Iterator<ZbnfParseResultItem> iterArguments = parent.iterChildren("argument");
      if(iterArguments != null)
      { while(iterArguments.hasNext())
        { ZbnfParseResultItem itemArgument = iterArguments.next();
          String sArg = gen_variableDefinition(itemArgument, null, null, 'a');
          sMethodHead += ", " + sArg;
        }
      }
    }
    sMethodHead += ", StacktraceThreadContext_s* _threadContext)";
    writeContent.writeClassH("METHOD_C " + sMethodHead + ";\n\n");

  }


  /**writes a methodDefinition with syntax methodDefinition::=.
   *
   * @param parent
   * @throws IOException
   * @throws ParseException
   */
  private void write_methodDeclaration(ZbnfParseResultItem parent, String sClassName)
  throws IOException, ParseException
  { ZbnfParseResultItem itemType = parent.getChild("typeIdent");
    String sDescription = get_description(parent);
    if(sDescription != null)
    { writeContent.writeClassH("/**" + sDescription + "*/\n");
    }
    String sType = gen_EnvIdent(itemType, "__");
    String sName = parent.getChild("name").getParsedString();
    String sMethodHead = sType + " " + sName + "_" + sClassName + "(" + sClassName + "* ythis";
    //
    Iterator<ZbnfParseResultItem> iterArguments = parent.iterChildren("argument");
    if(iterArguments != null)
    { while(iterArguments.hasNext())
      { ZbnfParseResultItem itemArgument = iterArguments.next();
        String sArg = gen_variableDefinition(itemArgument, null, null, 'a');
        sMethodHead += ", " + sArg;
      }
    }
    sMethodHead += ", StacktraceThreadContext_s* _threadContext)";
    writeContent.writeClassH("METHOD_C " + sMethodHead + ";\n\n");
  }



}
