/****************************************************************************
 * Copyright/Copyleft:
 *
 * For this source the LGPL Lesser General Public License,
 * published by the Free Software Foundation is valid.
 * It means:
 * 1) You can use this source without any restriction for any desired purpose.
 * 2) You can redistribute copies of this source to everybody.
 * 3) Every user of this source, also the user of redistribute copies
 *    with or without payment, must accept this license for further using.
 * 4) But the LPGL ist not appropriate for a whole software product,
 *    if this source is only a part of them. It means, the user
 *    must publish this part of source,
 *    but don't need to publish the whole source of the own product.
 * 5) You can study and modify (improve) this source
 *    for own using or for redistribution, but you have to license the
 *    modified sources likewise under this LGPL Lesser General Public License.
 *    You mustn't delete this Copyright/Copyleft inscription in this source file.
 *
 * @author JcHartmut: hartmut.schorrig@vishia.de
 * @version 2008-04-06  (year-month-day)
 * list of changes:
 * 2008-04-06 JcHartmut: some correction
 * 2008-03-15 JcHartmut: creation
 *
 ****************************************************************************/
package org.vishia.java2C;

import java.io.BufferedWriter;
import java.io.File;
import java.io.FileWriter;
import java.io.IOException;
import java.text.ParseException;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.TreeMap;

import org.vishia.zbnf.ZbnfParseResultItem;

/**This class processes the first and second pass of translating Java to C of a java file.
 * Parsing the content of the java file is done before, see {@link Java2C_Main}.
 * 
 * @author JcHartmut
 *
 */
public class GenerateFile implements iWriteContent
{

  /**Name without path and extension of the c- and h-file from java file. */
  final String sFileName;

  /**lines for c and h file mapping the class content. */
  private String sClassC = "" ;
  /**lines for c and h file mapping the class content. */
  private String sClassH = "" ;


  
  public void writeClassC(String line)
  { sClassC += line;
  }

  /**adds a line to headerfile string. */
  public void writeClassH(String line)
  { sClassH += line;
  }


  /**The interface to call a firstpass run of a needed type. */
  final RunRequiredFirstPass_ifc runRequiredFirstPass;


  /**All include declarations necessary in the C-file of this java file. 
   * The list will be filled also while running the second path.
   */
  private final TreeMap<String, String> includes = new TreeMap<String, String>();

  /**Adds a type to include the type-correspondent h-file in the c-file. */ 
  public void addIncludeC(String sType)
  { if(includes.get(sType) == null)
    { includes.put(sType, "c");  //not put if addIncludesH will be called before.
    }
  }

  /**Adds a type to include the type-correspondent h-file in the h-file. */ 
  public void addIncludeH(String sType)
  { includes.put(sType, "h");
  }



  private final TreeMap<String, String> imports = new TreeMap<String, String>();

  public String getImport(String sType)
  { return imports.get(sType);
  }




  /**List of all classes found in this file. This list is used to control, for which classes
   * the second path is to be run.
   */
  private final List<ClassData> itemsClassesForSecondPath = new LinkedList<ClassData>();

  /**initializes the instances with the given associations. */ 
  GenerateFile(String sFileName, RunRequiredFirstPass_ifc runRequiredFirstPass)
  { this.sFileName = sFileName;
    this.runRequiredFirstPass = runRequiredFirstPass;
  }


  /**Generates the content of a class in the h-file with given ZBNF parse result item of the class.
   * This routine is called from {@link Java2C_Main}.
   * <br>
   * First the h-file-content of all classes are generated into StringBuffer. If all runFirstPass
   * of all found classes are done, the file will be written. First the necessary includes will be written
   * in the destination h-file. Therefore the runFirstPass of all classes should be done before.
   *   
   * @param resultItem The first result item of ZBNF parsing of the whole java-File.
   * @param allClassData Global map of all classes of the whole translating process.
   * @param sPathOut output path for the h- and c-file, with slash on end
   * @throws IOException
   * @throws ParseException
   */
  void runFirstPass(ZbnfParseResultItem resultItem, TreeMap<String, ClassData> allClassData, String sPathOut)
  throws IOException, ParseException
  {
    //while(resultItem != null)
    gen_includes(resultItem);
    Iterator<ZbnfParseResultItem> iter= resultItem.iterChildren("classDefinition");
    while(iter != null && iter.hasNext())
    { resultItem = iter.next();
      String sSemantic = resultItem.getSemantic();
      if(sSemantic.equals("classDefinition") || sSemantic.equals("interfaceDefinition"))
      { FirstPass genClass = new FirstPass((iWriteContent)this, allClassData, runRequiredFirstPass);
        genClass.runFirstPass
        ( sFileName
        , resultItem   //the ZbnfParseResultItem of the class
        , ""
        , itemsClassesForSecondPath   //to add this class and sub classes if found
        );
      }
      //resultItem = resultItem.next(null);
    }
    iter= resultItem.iterChildren("interfaceDefinition");
    while(iter != null && iter.hasNext())
    { resultItem = iter.next();
      String sSemantic = resultItem.getSemantic();
      if(sSemantic.equals("classDefinition") || sSemantic.equals("interfaceDefinition"))
      { FirstPass genClass = new FirstPass((iWriteContent)this, allClassData, runRequiredFirstPass);
        genClass.runFirstPass
        ( sFileName
        , resultItem   //the ZbnfParseResultItem of the class
        , ""
        , itemsClassesForSecondPath   //to add this class and sub classes if found
        );
      }
      //resultItem = resultItem.next(null);
    }

    /**created header and c file. */
    BufferedWriter fileH; //, fileC;

    fileH = new BufferedWriter(new FileWriter(new File(sPathOut + sFileName + ".h")));
    //fileC = new BufferedWriter(new FileWriter(new File("gensrc_c/" + sFileName + ".c")));

    fileH.write("#ifndef __" + sFileName + "_h__\n");
    fileH.write("#define __" + sFileName + "_h__\n");
    fileH.write("//this file is generated by Java2C\n");
    fileH.write("\n#include \"ObjectJc.h\"\n");
    fileH.write("\n#include \"ExceptionJc.h\"\n");
    fileH.write("\n#include \"allReferences.h\"\n");

    //write includes:
    Set<Map.Entry<String, String>> listIncludeTypes = includes.entrySet();
    for(Map.Entry<String, String> entry: listIncludeTypes)
    { String sType = entry.getKey();
      String sValue = entry.getValue();
      if(sValue.equals("c"))
      { //fileC.write("#include \"" + sType + ".h\"\n");
      }
      else
      { fileH.write("#include \"" + sType + ".h\"\n");
      }
    }

    //fileH.write(sIncludeH);
    fileH.write(sClassH);
    fileH.write("#endif //__" + sFileName + "_h__\n");
    fileH.flush();

  }



  /**Generates the content of a class in the c-file. The class field {@link itemsClassesForSecondPath}
   * is filled while running the first pass and contains all classes found in the java file.
   * The ZBNF parse result item of each class is stored there.
   * This routine is called from {@link Java2C_Main}.
   * <br>
   * First the c-file-content of all classes are generated into the StringBuffer {@link sClassC}. 
   * If all runSecondPass of all found classes are done, the file will be written. 
   * First the necessary includes will be written
   * in the destination c-file. Therefore the runSecondPass of all classes should be done before.
   * @param allClassData Global map of all classes of the whole translating process.
   * @throws IOException
   * @throws ParseException
   */
  void runSecondPass(TreeMap<String, ClassData> allClassData) throws IOException, ParseException
  {


    sClassC = "";

    for(ClassData classData: itemsClassesForSecondPath)
    { ZbnfParseResultItem itemClass = classData.getParseResult();
      SecondPass genClass = new SecondPass((iWriteContent)this, allClassData, classData, runRequiredFirstPass);
      genClass.runSecondPass(itemClass, "");
      classData.relinquishParseResult();
    }


    /**created c file. */
    BufferedWriter fileC;
    fileC = new BufferedWriter(new FileWriter(new File("gensrc_c/" + sFileName + ".c")));
    fileC.write("//this file is generated by Java2C\n");
    fileC.write("#include \"" + sFileName + ".h\"\n");
    fileC.write("#include \"string.h\"  //because using memset()\n");
    fileC.flush();
    //write includes:
    Set<Map.Entry<String, String>> listIncludeTypes = includes.entrySet();
    for(Map.Entry<String, String> entry: listIncludeTypes)
    { String sType = entry.getKey();
      String sValue = entry.getValue();
      if(sValue.equals("c"))
      { fileC.write("#include \"" + sType + ".h\"\n");
      }
    }
    fileC.write(sClassC);
    fileC.flush();
  }


 
  /**generates include statements for the c-File from the import statements of the java file.
   * This includes may be dependencies to parts, there are contained in Java but not regarded
   * in Java2C translation. Such parts of Software may be supplied directly in C as alternativ
   * implementation of the Java parts. Especially the java parts may contain a emulation of environment
   * were the C parts contain the really environment adaption (drivers etc.).
   * @param resultFile The ZBNF parse result item of the whole java file, the < ?importStatements>
   *                   are getted from it.
   * @throws IOException
   */
  void gen_includes(ZbnfParseResultItem resultFile) throws IOException
  {
    List<ZbnfParseResultItem> itemsImport = resultFile.listChildren("importStatement");
    if(itemsImport!=null)for(ZbnfParseResultItem itemImport: itemsImport)
    { ZbnfParseResultItem importClass = itemImport.getChild("class");
      if(importClass != null)
      { String sImportClass = importClass.getParsedString();
        imports.put(sImportClass, "");
        addIncludeC(sImportClass);
      }
    }
  }



}
